{ IXF.PAS : Index formatter

  Title   : IXF
  Language: Borland Pascal v5.5 or higher.
  Version : 1.3
  Date    : Feb 15, 2000
  Author  : J R Ferguson
  Usage   : refer procedure Help
  Download: http://hello.to/ferguson
  E-mail  : j.r.ferguson@iname.com

This program and its source may be used and copied freely without charge,
but  only  for non-commercial purposes. The author is not responsible for
any damage or loss of data that may be caused by using it.

To compile this source file, you wil need  some  units  from  the  JRFPAS
Pascal  routine  library by the same author, which can be downloaded from
the Internet address mentioned above.
}

{$V-}                 { relaxed string parm checking }
{$S+}                 { stack overflow checking }
{$M 65520,0,655360}   { stack size, heap min, heap max }

program IXF;
uses DefLib, ArgLib, StpLib, ChrLib, Dos;

const
  LFTMARDFL = 5;  LFTMARMIN = 0; LFTMARMAX = 255; { output left margin }
  RGTMARDFL = 78; RGTMARMIN = 0; RGTMARMAX = 255; { output right margin }
  INPEXTDFL = '.IND';                             { default input filetype }
  OUTEXTDFL = '.IXF';                             { default output filetype }
  MAXFNM    = 79;         { Max filespec length (integrated environment) }
  IOBUFSIZ  = 32 * 1024;  { Input/Output buffer size in bytes }

  { Error codes and messages: }
  ERROK     = 0;    ERRARG    = 1;    ERRFNF    = 2;    ERRCRE    = 3;
  ERRREA    = 4;    ERRWRI    = 5;    ERRMEM    = 6;    ERRFMT    = 7;

  ERRMSG    : array[ERRFNF..ERRFMT] of StpTyp =
 ('Input file not found',
  'Output file creation error',
  'Read error',
  'Write error',
  'Out of memory',
  'Input file format error'
 );

type
  IOBufTyp  = array[1..IOBUFSIZ] of char;
  StpPtr    = ^StpTyp; { filled for actual length only }
  RefPtr    = ^RefRec; { linked list }
  NdxPtr    = ^NdxRec; { binary tree }
  RefRec    = record Ref: StpPtr; Nxt: RefPtr end;
  NdxRec    = record Ndx: StpPtr; Ref: RefPtr; Prv: NdxPtr; Nxt: NdxPtr end;

var
  ErrCod          : integer;
  InpFsp, OutFsp  : StpTyp;
  Inp, Out        : Text;
  InpOpn, OutOpn  : boolean;
  IOBuf           : IOBufTyp;
  NdxBuf          : NdxPtr;
  OptHlp          : boolean;
  LftMar          : StpInd;
  RgtMar          : StpInd;
  InpCnt, BufCnt  : longint;
  InpStp          : StpTyp;
  DotStp          : StpTyp;


{--- General routines ---}


procedure Help;
  procedure r(s: StpTyp);  begin write  (s) end;
  procedure i(i: StpInd);  begin write  (i) end;
  procedure l(s: StpTyp);  begin writeln(s) end;
begin
l('Usage  : IXF inpfile [outfile] [/option[...] [...]]');
l(' where : inpfile has default extent '+INPEXTDFL);
l('         outfile has default extent '+OUTEXTDFL+' and filename as in inpfile');
r('Options: Ln  Left  margin (n='); i(LFTMARMIN); r('..'); i(LFTMARMAX);
  r(', default '); i(LFTMARDFL); l(').');
r('         Rn  Right margin (n='); i(RGTMARMIN); r('..'); i(RGTMARMAX);
  r(', default '); i(RGTMARDFL); l('). n=0: no dot fill pattern.');
l('         H   Display full help');
if OptHlp then begin
l('');
l('Input file format (raw index):');
l('  comment lines: all spaces, or "*" in column 1');
l('  index entries: <spaces> <ref> <spaces> <text>');
l('Output file format (formatted index):');
l('  <margin> <text> ...................................... <ref>[,<ref>...]');
l('');
l('<spaces> = sequence of 1 or more space or tab characters.');
l('<ref>    = index entry (page) reference. No embedded spaces/tabs.');
l('<text>   = index entry text. May contain embedded spaces or tabs.');
l('<margin> = optional left margin (space characters).');
l('');
l('The Output file has the index text entries in lexicographic order (control,');
l('punctuation, digits, letters ignoring case).  For  lexicographically  equal');
l('entries the first text entry from the input file is copied, with all (page)');
l('references concatenated in the order found in the input file.');
end;
end;


{$F+} function HeapFunc(Size: word): integer; {$F-}
{ Make New() and GetMem() return a nil pointer when the heap is full }
begin HeapFunc:= 1; end;


function PrtLen(s: StpTyp): StpInd;
{ print length : recognizes backspace and skips other comtrol characters }
var i,l: StpInd;
begin
  l:= 0;
  for i:= 1 to Length(s) do begin
    if      IsPrint(s[i]) then Inc(l)
    else if (s[i] = chr(AsciiBS)) and (l>0) then Dec(l);
  end;
  PrtLen:= l;
end;


{--- Command line parsing routines ---}


procedure ParseFileSpecs;
var InpPth, InpFnm, InpExt: StpTyp;
    OutPth, OutFnm, OutExt: StpTyp;
begin
  InpPth:= ''; InpFnm:= ''; InpExt:= '';
  OutPth:= ''; OutFnm:= ''; OutExt:= '';
  FSplit(FExpand(InpFsp), InpPth, InpFnm, InpExt);
  FSplit(FExpand(OutFsp), OutPth, OutFnm, OutExt);
  if InpFnm = '' then ErrCod:= ERRARG
  else begin
    if OutFnm = '' then OutFnm:= InpFnm;
    if InpExt = '' then InpExt:= INPEXTDFL;
    if OutExt = '' then OutExt:= OUTEXTDFL;
    InpFsp:= InpPth + InpFnm + InpExt;
    OutFsp:= OutPth + OutFnm + OutExt;
  end;
end;


function ReadUns(var arg: StpTyp): integer;
var n: integer; c: char;
begin
  n:= 0; c:= StpcRet(arg,1);
  while (ErrCod=ERROK) and IsDigit(c) do begin
    StpDel(arg,1,1);
    n:= 10 * n + (ord(c) - ord('0'));
    c:= StpcRet(arg,1);
  end;
  ReadUns:= n;
end;


procedure ReadMargin(var arg: StpTyp; var margin: StpInd; min,max: StpInd);
var n: integer;
begin
  if not IsDigit(StpcRet(arg,1)) then ErrCod:= ERRARG
  else begin
    n:= ReadUns(arg);
    if (n < min) or (n > max) then ErrCod:= ERRARG
    else margin:= n;
  end;
end;


procedure ReadOpt(var arg: StpTyp);
var nextopt: boolean;
begin
  StpDel(arg,1,1);
  repeat
    if (arg = '') or (StpcRet(arg,1) = '/') then ErrCod:= ERRARG
    else begin
      nextopt:= false;
      while (ErrCod=ERROK) and not nextopt and (arg <> '') do
      case StpcGet(arg) of
        'L': ReadMargin(arg,LftMar,LFTMARMIN,LFTMARMAX);
        'R': ReadMargin(arg,RgtMar,RGTMARMIN,RGTMARMAX);
        'H': OptHlp:= true;
        '/': nextopt:= true;
        else ErrCod:= ERRARG;
      end;
    end;
  until (ErrCod <> ERROK) or not nextopt;
end;


procedure ReadArgs;
var i,f : ArgInd;
    arg : StpTyp;
begin
  GetArgs; i:= 0; f:= 0;
  while (i < ArgC) and (ErrCod = ERROK) do begin
    Inc(i); arg:= ArgV[i]; StpUpp(arg);
    case StpcRet(arg,1) of
      '/': ReadOpt(arg);
      else begin
             Inc(f);
             case f of
               1 : StpNCat(InpFsp,arg,MAXFNM);
               2 : StpNCat(OutFsp,arg,MAXFNM);
              else ErrCod:= ERRARG;
             end;
           end;
    end;
  end;
  if OptHlp then if ErrCod = ERROK then ErrCod:= ERRARG else OptHlp:= false;
  if ErrCod = ERROK then ParseFileSpecs;
end;


{--- Low-level I/O routines ---}


procedure OpenInp;
begin
  Assign(Inp,InpFsp); SetTextBuf(Inp,IOBuf);
  {$I-} reset(Inp); {$I+}
  if IOresult <> 0 then ErrCod:= ERRFNF else InpOpn:= true;
end;


procedure CloseInp;
begin if InpOpn then begin
  {$I-} Close(Inp); {$I+}
  if IOresult = 0 then InpOpn:= false;
end end;


function ReadInp(var line: StpTyp): boolean;
begin
  if (ErrCod <> ERROK) or eof(Inp) then ReadInp:= false
  else begin
    {$I-} ReadLn(Inp,line); {$I+}
    if IOResult <> 0 then begin ErrCod:= ERRREA; ReadInp:= false; end
    else ReadInp:= true;
  end;
end;


procedure OpenOut;
begin
  Assign(Out,OutFsp); SetTextBuf(Out,IOBuf);
  {$I-} rewrite(Out); {$I+}
  if IOresult <> 0 then ErrCod:= ERRCRE else OutOpn:= true;
end;


procedure CloseOut;
begin if OutOpn then begin
  {$I-} Close(Out); {$I+}
  if IOresult = 0 then OutOpn:= false else begin
    if ErrCod = ERROK then ErrCod:= ERRWRI;
  end;
end end;


procedure WriteOut(line: StpTyp);
begin
 {$I-} WriteLn(Out,line); {$I+}
 if IOResult <> 0 then ErrCod:= ERRWRI;
end;


{--- Index reference list handling routines (linked list: add at tail) ---}


procedure RefCreate(var pt: RefPtr; txt: StpTyp);
{ create new entry }
begin
  GetMem(pt,SizeOf(RefRec));
  if pt = nil then ErrCod:= ERRMEM
  else begin
    pt^.Nxt:= pt;
    GetMem(pt^.Ref,succ(Length(txt)));
    if pt^.Ref = nil then ErrCod:= ERRMEM
    else pt^.Ref^:= txt;
  end;
end;


procedure RefAdd(pt: RefPtr; txt: StpTyp);
{ add reference to wxisting entry }
var p: RefPtr;
begin
  p:= pt; while p^.Nxt <> pt do p:= p^.Nxt;
  if p^.Ref^ <> txt then begin
    RefCreate(p^.Nxt, txt);
    if ErrCod = ERROK then p^.Nxt^.Nxt:= pt;
  end;
end;


{--- Index buffer handling routines: binary tree (maintain in lexorder) ---}


procedure NdxInit(var ndx: NdxPtr);
{ initialise the structure }
begin ndx:= nil; end;


procedure NdxCreate(var ndx: NdxPtr; NdxStp, RefStp: StpTyp);
{ Create new node }
begin
  GetMem(ndx,SizeOf(NdxRec));
  if ndx = nil then ErrCod:= ERRMEM
  else begin
    ndx^.Prv:= nil; ndx^.Nxt:= nil;
    GetMem(ndx^.Ndx,succ(Length(NdxStp)));
    if ndx^.Ndx = nil then ErrCod:= ERRMEM
    else begin
      Inc(BufCnt); if (BufCnt mod 25) = 0 then write('.');
      ndx^.Ndx^:= NdxStp;
      RefCreate(ndx^.Ref,RefStp);
    end;
  end;
end;


procedure NdxInsert(var ndx: NdxPtr; NdxStp, RefStp: StpTyp);
{ Insert new node or add refenrence to exisiting node }

  var CmpVal: integer;

  procedure RecursiveNdxInsert(var ndx: NdxPtr);
  begin if ErrCod = ERROK then begin
    if ndx = nil then NdxCreate(ndx,NdxStp,RefStp)
    else begin
      CmpVal:= StpLexCmp(NdxStp,ndx^.Ndx^);
      if      CmpVal < 0 then RecursiveNdxInsert(ndx^.Prv)
      else if CmpVal = 0 then RefAdd(ndx^.Ref,RefStp)
      else if CmpVal > 0 then RecursiveNdxInsert(ndx^.Nxt);
    end;
  end end;

begin { NdxInsert }
  RecursiveNdxInsert(ndx);
end;


procedure NdxWrite(ndx: NdxPtr);
var line, NdxStp, RefStp: StpTyp; p: RefPtr; FillWidth: integer;

  procedure RecursiveNdxWrite(ndx: NdxPtr);
  begin if (ErrCod = ERROK) and (ndx <> nil) then begin
    RecursiveNdxWrite(ndx^.Prv);                           { left subtree }
    begin                                                  { current node }
      Inc(BufCnt); if (BufCnt mod 25) = 0 then write('.');
      NdxStp:= ndx^.Ndx^;
      p:= ndx^.Ref; RefStp:= p^.Ref^;
      while p^.Nxt <> ndx^.Ref do begin
        p:= p^.Nxt; RefStp:= RefStp + ',' + p^.Ref^;
      end;
      StpMake(line,' ',LftMar); line:= line + NdxStp + ' ';
      FillWidth:= RgtMar - PrtLen(line) - PrtLen(RefStp) - 1;
      if FillWidth > 1 then StpNCat(line,DotStp,FillWidth);
      line:= line + ' ' + RefStp; WriteOut(line);
    end;
    RecursiveNdxWrite(ndx^.Nxt);                           { right subtree }
  end; end;

begin { NdxWrite }
  RecursiveNdxWrite(ndx);
end;


{--- Processing routines ---}


procedure ReadText;
var NdxStp, RefStp: StpTyp;
begin
  OpenInp;
  if ErrCod = ERROK then begin
    write('Reading '); BufCnt:= 0;
    NdxInit(NdxBuf);
    while ReadInp(InpStp) do begin
      Inc(InpCnt); StpRTS(InpStp);
      if InpStp <> '' then if InpStp[1] <> '*' then begin
        NdxStp:= InpStp; StpGtw(RefStp,NdxStp); StpRLS(NdxStp);
        if NdxStp = '' then ErrCod:= ERRFMT
        else NdxInsert(NdxBuf,NdxStp,RefStp);
      end;
    end;
    CloseInp;
    writeln;
  end;
end;


procedure WriteText;
var p: NdxPtr;
begin
  OpenOut;
  if ErrCod = ERROK then begin
    write('Writing '); BufCnt:= 0;
    NdxWrite(NdxBuf);
    CloseOut;
    writeln;
  end;
end;


{--- Main line ---}


procedure MainProcess;
var Line: StpTyp; Heap: Pointer;
begin
  Mark(Heap);
  ReadText;
  if ErrCod = ERROK then begin InpCnt:= 0; WriteText; end;
  Release(Heap);
end;


procedure MainInit;
begin
  ErrCod:= ERROK; InpCnt:= 0; StpMake(DotStp,'.',MaxStp);
  HeapError:= @HeapFunc; {install function to catch Out-of-Heap condition}
  InpFsp:= ''; InpOpn:= false;
  OutFsp:= ''; OutOpn:= false;
  OptHlp:= false; LftMar:= LFTMARDFL; RgtMar:= RGTMARDFL;
  ReadArgs;
end;


procedure MainTerm;
begin
  if ErrCod = ERROK then writeln('Done')
  else begin
    if ErrCod=ERRARG then Help
    else writeln(ERRMSG[ErrCod]);
    if InpCnt > 0 then writeln('Line ',InpCnt,' ',InpStp);
  end;
end;


begin { Main program }
  writeln('IXF v1.3: Index formatter - J.R. Ferguson, Feb 2000');
  MainInit;
  if ErrCod = ERROK then begin
    writeln(InpFsp,' ==> ',OutFsp);
    MainProcess;
  end;
  MainTerm;
end.
